/**
 *
 * GLU Module for Python (incomplete)
 * 
 * Version: 0.85
 *
 * Date: Jan 98
 *
 * Authors: Thomas Schwaller  <tom.schwaller@linux-magazin.de>
 *          Jim Hugunin       <jjh@Goldilocks.LCS.MIT.EDU>
 *          David Ascher      <david_ascher@brown.edu>
 *          Michel Sanner     <sanner@scripps.edu>
 * 
 * Notes:
 *
 *   - the file gluconst.py is also needed, as it defines all kinds of
 *     useful constants
 *     
***/

#ifdef WIN32
#include <windows.h>
#define MCALLBACK (void (__stdcall *)(void))
#else
#define MCALLBACK
#endif

#include <GL/gl.h>
#include <GL/glu.h>
#include "Python.h"

#define ASSERT(E,M) if(!(E)) return ErrorReturn(M)
#define TRY(E) if(! (E)) return NULL
static int PyArray_AsDoubleArray(PyObject **op, GLdouble **pitems, int *pn)
{
    GLdouble *items;
    PyObject *item;
    int n, i;
    if (!PySequence_Check(*op)) return 0;
    n = PySequence_Length(*op);
    items = PyMem_NEW(GLdouble, n);
    if (items == NULL) {
	PyErr_NoMemory();
	return 0;
    }
    for (i=0; i<n; i++) {
	if (item = PySequence_GetItem(*op, i)) {
	   items[i] = PyFloat_AsDouble(item);
	   Py_DECREF(item);
        }
        if(PyErr_Occurred()) return 0;   
    }
    *pitems = items;
    *pn = n;
    return 1;
}

typedef struct {
  PyObject_HEAD
  GLUquadricObj* quad;
} gluQuadricObject;

typedef struct {
  PyObject_HEAD
  GLUtriangulatorObj* tess;
} gluTriangulatorObject;

typedef struct {
  PyObject_HEAD
  PyObject *pyfn;
  void (*fn)( );
} gluCallbackObject;

static PyTypeObject GLUquadricType;
static PyTypeObject GLUtriangulatorType;

#define is_gluQuadricObject(op) ((op)->ob_type == &GLUquadricType)
#define is_gluTriangulatorObject(op) ((op)->ob_type == &GLUtriangulatorType)
#define is_gluCallbackObject(op) ((op)->ob_type == &GLUcallbackType)

void
myVertex(GLenum arg)
{
	double *v;
	v = (double *) arg;
  glTexCoord2d(v[0], v[1]);
  glVertex3dv(v);
}


static PyObject* glu_Test(PyObject* self, PyObject* args)
{
  GLUtriangulatorObj *tobj;

  double v1[] = {0.0,1.0,1.0};
  double v2[] = {1.0,1.0,1.0};
  double v3[] = {1.0,0.0,1.0};
  double v4[] = {0.0,0.0,1.0};
  double v5[] = {0.2,0.2,1.0};
  double v6[] = {0.5,0.2,1.0};
  double v7[] = {0.2,0.5,1.0};
  
  tobj = gluNewTess();
  gluTessCallback(tobj, GLU_BEGIN, glBegin);
  gluTessCallback(tobj, GLU_VERTEX, MCALLBACK myVertex);
  gluTessCallback(tobj, GLU_END, glEnd);
  glColor3f(1.0,1.0,1.0);
  gluBeginPolygon(tobj);
  gluTessVertex(tobj, v1, v1);
  gluTessVertex(tobj, v2, v2);
  gluTessVertex(tobj, v3, v3);
  gluTessVertex(tobj, v4, v4);
  gluNextContour(tobj, GLU_INTERIOR); 
  gluTessVertex(tobj, v5, v5);
  gluTessVertex(tobj, v6, v6);
  gluTessVertex(tobj, v7, v7); 
  gluEndPolygon(tobj);

  Py_INCREF(Py_None);
  return Py_None;

}

/* QUADRIC TYPE STUFF */

GLUquadricObj* getgluquadricvalue(PyObject* op)
{
  if (!op || !is_gluQuadricObject(op)) {
    PyErr_BadInternalCall();
    return NULL;
  }
  else
    return ((gluQuadricObject*)op)->quad;    
}

PyObject* newgluQuadricObject(GLUquadricObj* Quad)
{
  gluQuadricObject* op;
  op = PyObject_NEW(gluQuadricObject, &GLUquadricType);
  if (op == NULL)
    return PyErr_NoMemory();
  op->ob_type = &GLUquadricType;
  op->quad = Quad;
  return (PyObject*)op;
}

static void gluquadric_delete(gluQuadricObject* op)
{
  if (op->quad) gluDeleteQuadric(op->quad);
  PyMem_DEL(op);
}

static PyObject* gluquadric_getattr(gluQuadricObject* op, char* name)
{
  static struct PyMethodDef glu_methods[] = {  
    {NULL,                   NULL}
  };
  return Py_FindMethod(glu_methods, (PyObject *) op, name);
}

static PyObject* glu_NewQuadric(PyObject* self, PyObject* args)
{
  if (!PyArg_ParseTuple(args, ""))
    return NULL;
  return newgluQuadricObject(gluNewQuadric());
}

/* TRIANGULATOR TYPE STUFF */

GLUtriangulatorObj* getglutriangulatorvalue(PyObject* op)
{
  if (!op || !is_gluTriangulatorObject(op)) {
    PyErr_BadInternalCall();
    return NULL;
  }
  else
    return ((gluTriangulatorObject*)op)->tess;    
}

PyObject* newgluTriangulatorObject(GLUtriangulatorObj* tess)
{
  gluTriangulatorObject* op;
  op = PyObject_NEW(gluTriangulatorObject, &GLUtriangulatorType);
  if (op == NULL)
    return PyErr_NoMemory();
  op->ob_type = &GLUtriangulatorType;
  op->tess = tess;

  return (PyObject*)op;
}


static void glutriangulator_delete(gluTriangulatorObject* op)
{
  if (op->tess) gluDeleteTess(op->tess);
  PyMem_DEL(op);
}

static PyObject* glutriangulator_getattr(gluTriangulatorObject* op, char* name)
{
  static struct PyMethodDef glu_methods[] = {  
    {NULL,                   NULL}
  };
  return Py_FindMethod(glu_methods, (PyObject *) op, name);
}

static PyObject* glu_NewTess(PyObject* self, PyObject* args)
{
  gluTriangulatorObject *o;
  GLUtriangulatorObj *tobj;
  tobj = gluNewTess();
  o = PyObject_NEW(gluTriangulatorObject, &GLUtriangulatorType);
  if (o == NULL)
    return PyErr_NoMemory();
  o->ob_type = &GLUtriangulatorType;
  o->tess = tobj;

  gluTessCallback(tobj, GLU_BEGIN, MCALLBACK glBegin);
  gluTessCallback(tobj, GLU_VERTEX, MCALLBACK myVertex);
  gluTessCallback(tobj, GLU_END, MCALLBACK glEnd);

  return (PyObject *)o;
}


/* CALLBACK TYPE STUFF */

/*
GLUcallbackObj* getglucallbackvalue(PyObject* op)
{
  if (!op || !is_gluCallbackObject(op)) {
    PyErr_BadInternalCall();
    return NULL;
  }
  else
    return ((gluCallbackObject*)op)->quad;    
}

PyObject* newgluCallbackObjectFromPyFun(GLUcallbackObj* Quad)
{
}

static void glucallback_delete(gluCallbackObject* op)
{
  if (op->quad) gluDeleteTess(op->quad);
  PyMem_DEL(op);
}

static PyObject* glucallback_getattr(gluCallbackObject* op, char* name)
{
  static struct PyMethodDef glu_methods[] = {  
    {NULL,                   NULL}
  };
  return Py_FindMethod(glu_methods, (PyObject *) op, name);
}

static PyObject* glu_CallbackFromPyFun(PyObject* self, PyObject* args)
{
  PyObject* pyfun;

  if (!PyArg_ParseTuple(args, "O"), &pyfun)
    return NULL;

  gluCallbackObject* op;
  op = PyObject_NEW(gluCallbackObject, &GLUcallbackType);
  if (op == NULL)
    return PyErr_NoMemory();
  op->ob_type = &GLUcallbackType;

  op->pyfn = pyfn;

  return (PyObject*)op;

}

static PyObject* glu_CallbackFromFun(PyObject* self, void (*fn)())
{
  gluCallbackObject* op;
  op = PyObject_NEW(gluCallbackObject, &GLUcallbackType);
  if (op == NULL)
    return PyErr_NoMemory();
  op->ob_type = &GLUcallbackType;

  op->fn = fn;

  return (PyObject*)op;
}
*/

static PyObject* glu_TessCallback(PyObject* self, PyObject* args)
{
  PyObject* tess;
  char *fun_name;
  GLenum which;

  if (!PyArg_ParseTuple(args, "Ois", &tess, &which, &fun_name))
    return NULL;

  if (strcmp(fun_name, "Begin") == 0) {
    gluTessCallback(getglutriangulatorvalue(tess), GLU_BEGIN, MCALLBACK glBegin);
  } else if (strcmp(fun_name, "End") == 0) {
    gluTessCallback(getglutriangulatorvalue(tess), GLU_VERTEX, glEnd);
  } else if (strcmp(fun_name, "Vertex") == 0) {
    gluTessCallback(getglutriangulatorvalue(tess), GLU_END, MCALLBACK glVertex3dv);
  }

  Py_INCREF(Py_None);
  return Py_None;
}


/* cylinder */
static PyObject* glu_Cylinder(PyObject* self, PyObject* args)
{

  PyObject* quad;
  GLdouble baseRadius, topRadius, height;
  GLint slices, stacks;
  if (!PyArg_ParseTuple(args, "Odddii", &quad, &baseRadius, &topRadius, &height, &slices, &stacks))
    return NULL;
  gluCylinder(getgluquadricvalue(quad), baseRadius, topRadius, height, slices, stacks);
  Py_INCREF(Py_None);
  return Py_None;
}

/* ---- spheres, partial disks */

static PyObject* glu_Sphere(PyObject* self, PyObject* args)
{

  PyObject* quad;
  GLdouble radius;
  GLint slices, stacks;
  if (!PyArg_ParseTuple(args, "Odii", &quad, &radius, &slices, &stacks))
    return NULL;
  gluSphere(getgluquadricvalue(quad), radius, slices, stacks);
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject* glu_PartialDisk(PyObject* self, PyObject* args)
{
  PyObject* quad;
  GLdouble iradius, oradius, start, sweep;
  GLint slices, loops;
  if (!PyArg_ParseTuple(args, "Oddiidd", &quad, &iradius, &oradius,
	&slices, &loops, &start, &sweep))
    return NULL;
  gluPartialDisk(getgluquadricvalue(quad), iradius, oradius,
		 slices, loops, start, sweep);
  Py_INCREF(Py_None);
  return Py_None;
}

/* Quadric draw style */

static PyObject *glu_QuadricDrawStyle(PyObject* self, PyObject* args)
{
  PyObject* quad;
  GLenum drawStyle;
  if(!PyArg_ParseTuple(args, "Oi", &quad, &drawStyle))
    return NULL;
  gluQuadricDrawStyle(getgluquadricvalue(quad), drawStyle);
  Py_INCREF(Py_None);
  return Py_None;
}

/* Quadric normals */

static PyObject* glu_QuadricNormals(PyObject* self, PyObject* args)
{
  PyObject* quad;
  GLenum normals;
  if(!PyArg_ParseTuple(args, "Oi", &quad, &normals))
    return NULL;
  gluQuadricNormals(getgluquadricvalue(quad), normals);
  Py_INCREF(Py_None);
  return Py_None;
}



static PyObject *glu_TessVertex(PyObject* self, PyObject* args)
{
  PyObject* tess;
  PyObject* d;
  double *vert;
  double *data;
  PyObject *v;
  int n;
  if(!PyArg_ParseTuple(args, "OOO", &tess, &v, &d))
    return NULL;
  TRY (PyArray_AsDoubleArray(&v, &vert, &n));
  TRY (PyArray_AsDoubleArray(&d, &data, &n));

  gluTessVertex(getglutriangulatorvalue(tess), vert, data);

  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject *glu_BeginPolygon(PyObject* self, PyObject* args)
{
  PyObject* tess;

  if(!PyArg_ParseTuple(args, "O", &tess))
    return NULL;
  gluBeginPolygon(getglutriangulatorvalue(tess));

  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject *glu_EndPolygon(PyObject* self, PyObject* args)
{
  PyObject* tess;
  
  if(!PyArg_ParseTuple(args, "O", &tess))
    return NULL;
  gluEndPolygon(getglutriangulatorvalue(tess));
  Py_INCREF(Py_None);
  return Py_None;
}


static PyObject *glu_NextContour(PyObject* self, PyObject* args)
{
  PyObject* tess;
  GLenum type;

  if(!PyArg_ParseTuple(args, "Oi", &tess, &type))
    return NULL;
  gluNextContour(getglutriangulatorvalue(tess), type);

  Py_INCREF(Py_None);
  return Py_None;
}

/* Perspectives etc. */

static PyObject *glu_Perspective(PyObject* self, PyObject* args)
{
  GLdouble a, b, c, d;
  if(!PyArg_ParseTuple(args, "dddd", &a, &b, &c, &d))
    return NULL;
  gluPerspective(a, b, c, d);
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject* glu_LookAt(PyObject* self, PyObject* args)
{
  GLdouble a, b, c, d, e, f, g, h, i;
  if(!PyArg_ParseTuple(args, "ddddddddd", &a, &b, &c, &d, &e, &f, &g, &h, &i))
    return NULL;
  gluLookAt(a, b, c, d, e, f, g, h, i);
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject* glu_Ortho2D(PyObject* self, PyObject* args)
{
  GLdouble a, b, c, d;
  if(!PyArg_ParseTuple(args, "dddd", &a, &b, &c, &d))
    return NULL;
  gluOrtho2D(a, b, c, d);
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject* glu_PickMatrix(PyObject* self, PyObject* args)
{
  int i;
  GLdouble x, y, width, height;
  GLint viewport[4];
  PyObject* viewportlist;
  if(!PyArg_ParseTuple(args, "ddddO!", &x, &y, &width, &height, 
                                       &PyList_Type, &viewportlist))
    return NULL;
  for (i=0; i<4; i++)
    viewport[i] = PyFloat_AsDouble(PyList_GetItem(viewportlist, i));
  gluPickMatrix(x, y, width, height, viewport);
  Py_INCREF(Py_None);
  return Py_None;  
}

static PyObject* glu_UnProject(PyObject* self, PyObject* args)
{
   GLdouble winx, winy, winz;
   GLdouble objx, objy, objz;
   GLdouble model[16], proj[16];
   GLint view[16];

   if(!PyArg_ParseTuple(args, "ddd", &winx, &winy, &winz))
      return NULL;

   glGetDoublev(GL_MODELVIEW_MATRIX, model);
   glGetDoublev(GL_PROJECTION_MATRIX, proj);
   glGetIntegerv(GL_VIEWPORT, view);
   gluUnProject(winx, winy, winz, model, proj, view, &objx, &objy, &objz);

   return Py_BuildValue("(ddd)", objx, objy, objz);
}

static PyObject* glu_Project(PyObject* self, PyObject* args)
{
   GLdouble winx, winy, winz;
   GLdouble objx, objy, objz;
   GLdouble model[16], proj[16];
   GLint view[16];

   if(!PyArg_ParseTuple(args, "ddd", &objx, &objy, &objz))
      return NULL;

   glGetDoublev(GL_MODELVIEW_MATRIX, model);
   glGetDoublev(GL_PROJECTION_MATRIX, proj);
   glGetIntegerv(GL_VIEWPORT, view);
   gluProject(objx, objy, objz, model, proj, view, &winx, &winy, &winz);

   return Py_BuildValue("(ddd)", winx, winy, winz);
}

/* MS Dec 12 '97 */
static PyObject *glu_ErrorString(PyObject *self, PyObject *args)
{
  int errCode;

  if (! PyArg_ParseTuple(args, "i", &errCode) )
    return NULL;
  return Py_BuildValue("s", gluErrorString((GLenum)errCode));
}

/* End MS Dec 12 '97 */


static PyTypeObject GLUquadricType = {
#ifdef WIN32   
	PyObject_HEAD_INIT(NULL)  
#else
	PyObject_HEAD_INIT(&PyType_Type)  
#endif
  0,			           /* Number of items for varobject */
  "glu",                           /* Name of this type */
  sizeof(gluQuadricObject),        /* Basic object size */
  0,			           /* Item size for varobject */
  (destructor)gluquadric_delete,   /* tp_dealloc*/
  0,                               /* tp_print*/
  (getattrfunc)gluquadric_getattr, /* tp_getattr*/
  0,			           /* tp_setattr*/
  0,                               /* tp_compare*/
  0,                               /* tp_repr*/
  0,                               /* tp_as_number*/
  0,	                           /* tp_as_sequence*/
  0,			           /* tp_as_mapping*/
};



static PyTypeObject GLUtriangulatorType = {
#ifdef WIN32   
	PyObject_HEAD_INIT(NULL)  
#else
	PyObject_HEAD_INIT(&PyType_Type)  
#endif
  0,			           /* Number of items for varobject */
  "glu",                           /* Name of this type */
  sizeof(gluTriangulatorObject),        /* Basic object size */
  0,			           /* Item size for varobject */
  (destructor)glutriangulator_delete,   /* tp_dealloc*/
  0,                               /* tp_print*/
  (getattrfunc)glutriangulator_getattr, /* tp_getattr*/
  0,			           /* tp_setattr*/
  0,                               /* tp_compare*/
  0,                               /* tp_repr*/
  0,                               /* tp_as_number*/
  0,	                           /* tp_as_sequence*/
  0,			           /* tp_as_mapping*/
};

static PyMethodDef glu_methods[] = {
  { "gluNewQuadric",        glu_NewQuadric , 1},
  { "gluNewTess",           glu_NewTess, 1},
  { "gluSphere",            glu_Sphere , 1},  
  { "gluCylinder",          glu_Cylinder, 1},
  { "gluPartialDisk",       glu_PartialDisk , 1},  
  { "gluQuadricDrawStyle",  glu_QuadricDrawStyle , 1},    
  { "gluQuadricNormals",    glu_QuadricNormals , 1},    
  { "gluPerspective",       glu_Perspective, 1}, 
  { "gluLookAt",            glu_LookAt, 1},   
  { "gluOrtho2D",           glu_Ortho2D, 1},    
  { "gluPickMatrix",        glu_PickMatrix, 1},  
  { "gluUnProject",         glu_UnProject, 1},
  { "gluProject",           glu_Project, 1},
  { "gluTessVertex",        glu_TessVertex, 1},
  { "gluBeginPolygon",      glu_BeginPolygon, 1},
  { "gluEndPolygon",        glu_EndPolygon, 1},
  { "gluNextContour",       glu_NextContour, 1},
  { "gluTessCallback",      glu_TessCallback, 1},
  { "gluTest",      glu_Test, 1},
  { "gluErrorString",      glu_ErrorString, 1},
  { NULL,                NULL }
};

void init_glu()
{
#ifdef WIN32
	GLUtriangulatorType.ob_type = &PyType_Type;
	GLUquadricType.ob_type = &PyType_Type;
#endif
	Py_InitModule("_glu", glu_methods);
}

