/* $Id: _numarraymodule.c,v 1.5 2002/04/24 13:09:01 jaytmiller Exp $ */

#include <Python.h>
#include <stdio.h>
#include <math.h>
#include <signal.h>

#include "libnumarray.h"

typedef void Sigfunc(int);

static PyObject *_numarrayError;

/* The following function has much platform dependent code since
** there is no platform-independent way of checking Floating Point
** status bits
*/

static PyObject *CheckFPErrors(PyObject *self, PyObject *args) {

#ifdef __osf__
#ifdef __alpha
    unsigned long fpstatus;
#endif
#else
    int fpstatus;
#endif
    int retstatus;
/* MS Windows -----------------------------------------------------*/
#if defined(_MSC_VER)
#include <float.h>
    fpstatus = (int) _clear87();
    retstatus = pyFPE_DIVIDE_BY_ZERO * ((SW_ZERODIVIDE & fpstatus) != 0)
              + pyFPE_OVERFLOW       * ((SW_OVERFLOW & fpstatus)   != 0)
              + pyFPE_UNDERFLOW      * ((SW_UNDERFLOW & fpstatus)  != 0)
              + pyFPE_INVALID        * ((SW_INVALID & fpstatus)    != 0);

/* Solaris --------------------------------------------------------*/
/* --------ignoring SunOS ieee_flags approach, someone else can
**         deal with that! */
#elif defined(sun)
#include <ieeefp.h>
/* #include <math.h> */
    fpstatus = (int) fpgetsticky();
    retstatus = pyFPE_DIVIDE_BY_ZERO * ((FP_X_DZ  & fpstatus) != 0)
              + pyFPE_OVERFLOW       * ((FP_X_OFL & fpstatus) != 0)
              + pyFPE_UNDERFLOW      * ((FP_X_UFL & fpstatus) != 0)
              + pyFPE_INVALID        * ((FP_X_INV & fpstatus) != 0);
    (void) fpsetsticky(0);

/* Linux ----------------------------------------------------------*/
#elif defined(linux)
#ifdef __GLIBC__
#include <fenv.h>
#else
#include <i386/fenv.h>
#endif
    fpstatus = (int) fetestexcept(
                   FE_DIVBYZERO | FE_OVERFLOW | FE_UNDERFLOW | FE_INVALID);
    retstatus = pyFPE_DIVIDE_BY_ZERO * ((FE_DIVBYZERO  & fpstatus) != 0)
              + pyFPE_OVERFLOW       * ((FE_OVERFLOW   & fpstatus) != 0)
              + pyFPE_UNDERFLOW      * ((FE_UNDERFLOW  & fpstatus) != 0)
              + pyFPE_INVALID        * ((FE_INVALID    & fpstatus) != 0);
    (void) feclearexcept(FE_DIVBYZERO | FE_OVERFLOW | FE_UNDERFLOW | FE_INVALID);

/* Tru64 ----------------------------------------------------------*/
#elif defined(__alpha)
#ifdef __osf__
#include <machine/fpu.h>
    fpstatus = ieee_get_fp_control(); 
    /* clear status bits as well as disable exception mode if on */
    ieee_set_fp_control( 0 );
    retstatus = pyFPE_DIVIDE_BY_ZERO* (int)((IEEE_STATUS_DZE  & fpstatus) != 0)
              + pyFPE_OVERFLOW      * (int)((IEEE_STATUS_OVF  & fpstatus) != 0)
              + pyFPE_UNDERFLOW     * (int)((IEEE_STATUS_UNF  & fpstatus) != 0)
              + pyFPE_INVALID       * (int)((IEEE_STATUS_INV  & fpstatus) != 0); 
#endif
#else
    retstatus = 0;
#endif
    return Py_BuildValue("i", retstatus);
}

/* fill the buffer with values = start + index*delta */

static int fillarrayInt32 (long niter, long ninargs, long noutargs, 
			   void **buffers, long *bsizes) {
    long i, start, delta;
    Int32 *tin, *tparams;
    Int8 itemsizes[2] = { sizeof(Int32), sizeof(Int32) };
    Int8 iters[2] = { 0, 2 };

    if (NA_checkIo("fillarrayInt32", 1, 1, ninargs, noutargs) ||
	NA_checkNCBuffers("fillarrayInt32", 2,
			 niter, buffers, bsizes, itemsizes, iters))
	    return -1;

    tin = (Int32 *) buffers[0];
    tparams = (Int32 *) buffers[1];
    start = tparams[0];
    delta = tparams[1];

    for (i=0; i<niter; i++) {
        *(tin++) = start;
        start += delta;
    }
    return 0;
}

SELF_CHECKED_CFUNC_DESCR(fillarrayInt32, CFUNC_UFUNC);

static int fillarrayFloat32 (long niter, long ninargs, long noutargs, 
			     void **buffers, long *bsizes) {
    long i;
    double start, delta;
    Float32 *tin, *tparams;
    Int8 itemsizes[2] = { sizeof(Float32), sizeof(Float32) };
    Int8 iters[2] = { 0, 2 };

    if (NA_checkIo("fillarrayFloat32", 1, 1, ninargs, noutargs) ||
	NA_checkNCBuffers("fillarrayFloat32", 2,
			 niter, buffers, bsizes, itemsizes, iters))
	    return -1;

    tin = (Float32 *) buffers[0];
    tparams = (Float32 *) buffers[1];
    start = tparams[0];
    delta = tparams[1];

    for (i=0; i<niter; i++) {
        *(tin++) = (Float32) start;
        start += delta;
    }
    return 0;
}

SELF_CHECKED_CFUNC_DESCR(fillarrayFloat32, CFUNC_UFUNC);

static int fillarrayFloat64 (long niter, long ninargs, long noutargs, 
			     void **buffers, long *bsizes) {
    long i;
    double start, delta;
    Float64 *tin, *tparams;
    Int8 itemsizes[2] = { sizeof(Float64), sizeof(Float64) };
    Int8 iters[2] = { 0, 2 };

    if (NA_checkIo("fillarrayFloat64", 1, 1, ninargs, noutargs) ||
	NA_checkNCBuffers("fillarrayFloat64", 2,
			 niter, buffers, bsizes, itemsizes, iters))
	    return -1;

    tin = (Float64 *) buffers[0];
    tparams = (Float64 *) buffers[1];
    start = tparams[0];
    delta = tparams[1];

    for (i=0; i<niter; i++) {
        *(tin++) = start;
        start += delta;
    }
    return 0;
}

SELF_CHECKED_CFUNC_DESCR(fillarrayFloat64, CFUNC_UFUNC);


static int nonzeroCoordsBool(long niter, long ninargs, long noutargs,
                               void **buffers, long *bsizes)
{
	Bool   *source;
	Int32  *sstride;
	Int32  **output;
	Int32 i, minbsize, next = 0;
	
	if (NA_checkIo("nonzeroCoordsBool", 2, noutargs, ninargs, noutargs))
		return -1;

	if (NA_checkOneCBuffer("nonzeroCoordsBool", niter,
		      buffers[0], bsizes[0], sizeof(Bool)))
		return -1;
	    
	if (NA_checkOneCBuffer("nonzeroCoordsBool", noutargs,
		      buffers[1], bsizes[1], sizeof(Int32)))
		return -1;

	/* Check alignment only.  Storage optimization makes arrays
	 shorter than niter likely, possibly even 0! */
	for (i=0; i<noutargs; i++)
		if (NA_checkOneCBuffer("nonzeroCoordsBool", 0,
			      buffers[2+i], bsizes[2+i], sizeof(Int32)))
			return -1;

	source     = (Int8 *)    buffers[0];
	sstride    = (Int32  *)  buffers[1];
	output     = (Int32 **) &buffers[2];

	minbsize = bsizes[0];
	for(i=0; i<noutargs; i++)
		if (sstride[i] <= 0) {
			PyErr_Format(_numarrayError,
			     "nonzeroCoordsBool: bad stride[%d] = %d.\n",
				     i, sstride[i]);
			return -1;
		}
		else if (bsizes[i+2] < minbsize)
			minbsize = bsizes[i+2];
	
	next = 0;
	for(i=0; i<niter; i++)
	{
		if (source[i] != 0)
		{
			Int32 j, index = i;
			if (next >= minbsize) {
				PyErr_Format(_numarrayError,
					     "nonzeroCoordsBool: insufficient index space.\n");
				return -1;
			}
			for(j=0; j<noutargs; j++)
			{
				output[j][next] = index / sstride[j];
				index %= sstride[j];
			}
			++ next;
		}
	}
	return 0;
}

SELF_CHECKED_CFUNC_DESCR(nonzeroCoordsBool, CFUNC_UFUNC);

static int mxmultiplyInt32( long niter, long ninargs, long noutargs,
			       void **buffers, long *bsizes)
{
  Int32 *a, *b;
  Int32 maxi, maxj, maxk;
  Int32 *r;
  int i, j, k;

  if (NA_checkIo("mxmultiplyInt32", 3, 1, ninargs, noutargs))
	  return -1;

  if (NA_checkOneCBuffer("mxmultiplyInt32", 1,
		buffers[2], bsizes[2], sizeof(Int32)))
	  return -1;
  else {
	  maxi = ((Int32*)buffers[2])[0];
	  maxj = ((Int32*)buffers[2])[1];
	  maxk = ((Int32*)buffers[2])[2];
  }
	  
  if (NA_checkOneCBuffer("mxmultiplyInt32", maxi*maxk,
		buffers[0], bsizes[0], sizeof(Int32)))
	  return -1;
  else
	  a   = (Int32 *) buffers[0];

  if (NA_checkOneCBuffer("mxmultiplyInt32", maxj*maxk,
		buffers[1], bsizes[1], sizeof(Int32)))
	  return -1;
  else
	  b   = (Int32 *) buffers[1];

  if (NA_checkOneCBuffer("mxmultiplyInt32", maxi*maxj,
		buffers[3], bsizes[3], sizeof(Int32)))
	  return -1;
  else
	  r   = (Int32 *) buffers[3];

  for(i=0; i<maxi; i++)
    for(j=0; j<maxj; j++)
      {
	Int32 s = 0;
	Int32 *ap = a + i*maxk;
	Int32 *bp = b + j*maxk;
	for(k=0; k<maxk; k++)
	  s += *ap++ * *bp++;
	*r++ = s;
      }

  return 0;
}

SELF_CHECKED_CFUNC_DESCR(mxmultiplyInt32, CFUNC_UFUNC);

static int mxmultiplyFloat64( long niter, long ninargs, long noutargs,
			       void **buffers, long *bsizes)
{
  Float64 *a, *b;
  Int32   maxi, maxj, maxk;
  Float64 *r;
  int i, j, k;

  if (NA_checkIo("mxmultiplyFloat64", 3, 1, ninargs, noutargs))
	  return -1;

  if (NA_checkOneCBuffer("mxmultiplyFloat64", 1,
		buffers[2], bsizes[2], sizeof(Int32)))
	  return -1;
  else {
	  maxi = ((Int32*)buffers[2])[0];
	  maxj = ((Int32*)buffers[2])[1];
	  maxk = ((Int32*)buffers[2])[2];
  }
	  
  if (NA_checkOneCBuffer("mxmultiplyFloat64", maxi*maxk,
		buffers[0], bsizes[0], sizeof(Float64)))
	  return -1;
  else
	  a   = (Float64 *) buffers[0];

  if (NA_checkOneCBuffer("mxmultiplyFloat64", maxj*maxk,
		buffers[1], bsizes[1], sizeof(Float64)))
	  return -1;
  else
	  b   = (Float64 *) buffers[1];

  if (NA_checkOneCBuffer("mxmultiplyFloat64", maxi*maxj,
		buffers[3], bsizes[3], sizeof(Float64)))
	  return -1;
  else
	  r   = (Float64 *) buffers[3];

  for(i=0; i<maxi; i++)
    for(j=0; j<maxj; j++)
      {
	Float64 s = 0;
	Float64 *ap = a + i*maxk;
	Float64 *bp = b + j*maxk;
	for(k=0; k<maxk; k++)		
	  s += *ap++ * *bp++;
	*r++ = s;
      }
  return 0;
}

SELF_CHECKED_CFUNC_DESCR(mxmultiplyFloat64, CFUNC_UFUNC);

static int mxmultiplyComplex64( long niter, long ninargs, long noutargs,
			       void **buffers, long *bsizes)
{
  Complex64 *a, *b;
  Int32   maxi, maxj, maxk;
  Complex64 *r;
  int i, j, k;

  if (NA_checkIo("mxmultiplyComplex64", 3, 1, ninargs, noutargs))
	  return -1;

  if (NA_checkOneCBuffer("mxmultiplyComplex64", 1,
		buffers[2], bsizes[2], sizeof(Int32)))
	  return -1;
  else {
	  maxi = ((Int32*)buffers[2])[0];
	  maxj = ((Int32*)buffers[2])[1];
	  maxk = ((Int32*)buffers[2])[2];
  }
	  
  if (NA_checkOneCBuffer("mxmultiplyComplex64", maxi*maxk,
		buffers[0], bsizes[0], sizeof(Complex64)))
	  return -1;
  else
	  a   = (Complex64 *) buffers[0];

  if (NA_checkOneCBuffer("mxmultiplyComplex64", maxj*maxk,
		buffers[1], bsizes[1], sizeof(Complex64)))
	  return -1;
  else
	  b   = (Complex64 *) buffers[1];

  if (NA_checkOneCBuffer("mxmultiplyComplex64", maxi*maxj,
		buffers[3], bsizes[3], sizeof(Complex64)))
	  return -1;
  else
	  r   = (Complex64 *) buffers[3];

  for(i=0; i<maxi; i++)
    for(j=0; j<maxj; j++)
      {
	Complex64 s = { 0, 0 };
	Complex64 *ap = a + i*maxk;
	Complex64 *bp = b + j*maxk;
	for(k=0; k<maxk; k++, ap++, bp++) {
		Complex64 t;
		NUM_CMUL(*ap, *bp, t);
		NUM_CADD(s, t, s);
	}
	*r++ = s;
      }
  return 0;
}

SELF_CHECKED_CFUNC_DESCR(mxmultiplyComplex64, CFUNC_UFUNC);

static PyMethodDef _numarrayMethods[] = {
    {"CheckFPErrors", CheckFPErrors, METH_VARARGS},
    {NULL,      NULL}        /* Sentinel */
};

static PyObject *init_numarrayDict(void) {
    PyObject *dict;
    int status;
    dict = PyDict_New();
    status = NA_add_cfunc(dict, "fillarrayInt32",    
		     (void *) &fillarrayInt32_descr);
    status = NA_add_cfunc(dict, "fillarrayFloat32",  
		     (void *) &fillarrayFloat32_descr);
    status = NA_add_cfunc(dict, "fillarrayFloat64",  
		     (void *) &fillarrayFloat64_descr);
    status = NA_add_cfunc(dict, "nonzeroCoordsBool", 
		     (void *) &nonzeroCoordsBool_descr);
    status = NA_add_cfunc(dict, "mxmultiplyInt32",   
		     (void *) &mxmultiplyInt32_descr);
    status = NA_add_cfunc(dict, "mxmultiplyFloat64", 
		     (void *) &mxmultiplyFloat64_descr);
    status = NA_add_cfunc(dict, "mxmultiplyComplex64", 
		     (void *) &mxmultiplyComplex64_descr);
    return dict;
}

/* platform independent*/
#ifdef MS_WIN32
__declspec(dllexport)
#endif

void init_numarray(void) {
    PyObject *m, *d;
    m = Py_InitModule("_numarray", _numarrayMethods);
    d = PyModule_GetDict(m);
    _numarrayError = PyErr_NewException("_numarray.error", NULL, NULL);
    PyDict_SetItemString(d, "error", _numarrayError);
    import_libnumarray();
    PyDict_SetItemString(d, "functionDict", init_numarrayDict());
}
