/*
* Simple interface to libxslt
*
*   Copyright (c) 2001 Dave Kuhlman
*   
*   Permission is hereby granted, free of charge, to any person obtaining
*   a copy of this software and associated documentation files (the
*   "Software"), to deal in the Software without restriction, including
*   without limitation the rights to use, copy, modify, merge, publish,
*   distribute, sublicense, and/or sell copies of the Software, and to
*   permit persons to whom the Software is furnished to do so, subject to
*   the following conditions:
*   
*   The above copyright notice and this permission notice shall be
*   included in all copies or substantial portions of the Software.
*   
*   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
*   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
*   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
*   IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
*   CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
*   TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
*   SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*
*/

/* Implementation of Python extension datatype libxslt_stylesheet.
*/


/* Libxslt_stylesheet objects */

#include "libxml/xmlmemory.h"
#include "libxml/debugXML.h"
#include "libxml/HTMLtree.h"
#include "libxml/xmlIO.h"
#include "libxml/DOCBparser.h"
#include "libxml/xinclude.h"
#include "libxml/catalog.h"
#include "libxslt/xslt.h"
#include "libxslt/xsltInternals.h"
#include "libxslt/transform.h"
#include "libxslt/xsltutils.h"

/* Use this file as a template to start implementing a module that
   also declares object types. All occurrences of 'Xxo' should be changed
   to something reasonable for your objects. After that, all other
   occurrences of 'xx' should be changed to something reasonable for your
   module. If your module is named foo your sourcefile should be named
   foomodule.c.
   
   You will probably want to delete all references to 'x_attr' and add
   your own types of attributes instead.  Maybe you want to name your
   local variables other than 'self'.  If your object type is needed in
   other files, you'll have to create a file "foobarobject.h"; see
   intobject.h for an example. */

/* Xxo objects */

#include "Python.h"

#include "libxslt_stylesheet.h"

staticforward PyTypeObject Libxslt_stylesheet_Type;

#define Libxslt_stylesheetObject_Check(v)	((v)->ob_type == &Libxslt_stylesheet_Type)


Libxslt_stylesheetObject *
newLibxslt_stylesheetObject(xsltStylesheetPtr underlyingObject)
{
	Libxslt_stylesheetObject *self;
	self = PyObject_New(Libxslt_stylesheetObject, &Libxslt_stylesheet_Type);
	if (self == NULL)
	{
		return NULL;
	} /* if */
	self->x_attr = NULL;
	self->obj_xsltStylesheetPtr = underlyingObject;
	return self;
} /* newLibxslt_stylesheetObject */

/* Libxslt_stylesheet methods */

static void
Libxslt_stylesheet_dealloc(Libxslt_stylesheetObject *self)
{
	Py_XDECREF(self->x_attr);
	xsltFreeStylesheet(self->obj_xsltStylesheetPtr);
	PyObject_Del(self);
} /* Libxslt_stylesheet_dealloc */


static PyMethodDef Libxslt_stylesheet_methods[] = {
	{NULL,		NULL}		/* sentinel */
};

static PyObject *
Libxslt_stylesheet_getattr(Libxslt_stylesheetObject *self, char *name)
{
	if (self->x_attr != NULL)
	{
		PyObject *v = PyDict_GetItemString(self->x_attr, name);
		if (v != NULL)
		{
			Py_INCREF(v);
			return v;
		} /* if */
	} /* if */
	return Py_FindMethod(Libxslt_stylesheet_methods, (PyObject *)self, name);
} /* Libxslt_stylesheet_getattr */

static int
Libxslt_stylesheet_setattr(Libxslt_stylesheetObject *self, char *name, PyObject *v)
{
	if (self->x_attr == NULL)
	{
		self->x_attr = PyDict_New();
		if (self->x_attr == NULL)
		{
			return -1;
		} /* if */
	} /* if */
	if (v == NULL)
	{
		int rv = PyDict_DelItemString(self->x_attr, name);
		if (rv < 0)
		{
			PyErr_SetString(PyExc_AttributeError,
			        "delete non-existing Libxslt_stylesheet attribute");
		} /* if */
		return rv;
	}
	else
	{
		return PyDict_SetItemString(self->x_attr, name, v);
	} /* if */
} /* Libxslt_stylesheet_setattr */

statichere PyTypeObject Libxslt_stylesheet_Type = {
	/* The ob_type field must be initialized in the module init function
	 * to be portable to Windows without using C++. */
	PyObject_HEAD_INIT(NULL)
	0,			/*ob_size*/
	LIBXSLT_STYLESHEET_NAME,		/*tp_name*/
	sizeof(Libxslt_stylesheetObject),	/*tp_basicsize*/
	0,			/*tp_itemsize*/
	/* methods */
	(destructor)Libxslt_stylesheet_dealloc, /*tp_dealloc*/
	0,			/*tp_print*/
	(getattrfunc)Libxslt_stylesheet_getattr, /*tp_getattr*/
	(setattrfunc)Libxslt_stylesheet_setattr, /*tp_setattr*/
	0,			/*tp_compare*/
	0,			/*tp_repr*/
	0,			/*tp_as_number*/
	0,			/*tp_as_sequence*/
	0,			/*tp_as_mapping*/
	0,			/*tp_hash*/
	0,			/*tp_call*/
	0,			/*tp_str*/
	0,			/*tp_getattro*/
	0,			/*tp_setattro*/
	0,			/*tp_as_buffer*/
	Py_TPFLAGS_DEFAULT,     /*tp_flags*/
	0,			/*tp_doc*/
	0,			/*tp_traverse*/
	0,			/*tp_clear*/
	0,			/*tp_richcompare*/
	0,			/*tp_weaklistoffset*/
	0,			/*tp_iter*/
	0,			/*tp_iternext*/
	0,			/*tp_methods*/
	0,			/*tp_members*/
	0,			/*tp_getset*/
	0,			/*tp_base*/
	0,			/*tp_dict*/
	0,			/*tp_descr_get*/
	0,			/*tp_descr_set*/
	0,			/*tp_dictoffset*/
	0,			/*tp_init*/
	0,			/*tp_alloc*/
	0,			/*tp_new*/
	0,			/*tp_free*/
	0,			/*tp_is_gc*/
};
/* --------------------------------------------------------------------- */

